from utils import *
import numpy as np
import math
from ._dist_struct_robot_auto import DistStructRobotAuto
from ._dist_struct_robot_auto import DistStructLinkLink
import os
if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp

def _diststructlinklink_cpp2py(dsll_cpp):

    dsll_py = DistStructLinkLink(dsll_cpp.link_number_1, dsll_cpp.link_col_obj_number_1, dsll_cpp.link_number_2, dsll_cpp.link_col_obj_number_2, dsll_cpp.distance, np.matrix(dsll_cpp.point_link_1).reshape((3,1)), np.matrix(dsll_cpp.point_link_2).reshape((3,1)), np.matrix(dsll_cpp.jac_distance))
    return dsll_py

def _diststructlinklink_py2cpp(dsll_py):

    dsll_cpp = ub_cpp.CPP_DistStructLinkLink()

    dsll_cpp.is_null = False
    dsll_cpp.link_number_1 = dsll_py.link_number_1
    dsll_cpp.link_number_2 = dsll_py.link_number_2
    dsll_cpp.link_col_obj_number_1 = dsll_py.link_col_obj_number_1
    dsll_cpp.link_col_obj_number_2 = dsll_py.link_col_obj_number_2
    dsll_cpp.distance = dsll_py.distance
    dsll_cpp.point_link_1 = dsll_py.point_link_1
    dsll_cpp.point_link_2 = dsll_py.point_link_2
    dsll_cpp.jac_distance = dsll_py.jac_distance

    return dsll_cpp

def _diststructrobotauto_cpp2py(dsra_cpp, _robot):

    dsra_py = DistStructRobotAuto(_robot)

    dsra_py._jac_dist_mat = np.matrix(dsra_cpp.jac_dist_mat)
    n = np.shape(dsra_cpp.dist_vect)[0]
    dsra_py._dist_vect = np.matrix(dsra_cpp.dist_vect).reshape((n,1))
    
    dsra_py._list_info = [_diststructlinklink_cpp2py(dsll) for dsll in dsra_cpp.list_info]
    dsra_py._no_items = len(dsra_py._list_info)

    return dsra_py

def _diststructrobotauto_py2cpp(dsra_py):

    dsra_cpp = ub_cpp.CPP_DistStructRobotAuto()

    dsra_cpp.is_null = dsra_py is None
    if not (dsra_py is None):
        dsra_cpp.jac_dist_mat = dsra_py.jac_dist_mat
        dsra_cpp.dist_vect = dsra_py.dist_vect
        dsra_cpp.list_info = [_diststructlinklink_py2cpp(dsll) for dsll in dsra_py._list_info]

    return dsra_cpp


def _compute_dist_auto(self, q=None, old_dist_struct=None, tol=0.0005, no_iter_max=20, max_dist = np.inf, h=0, eps = 0, mode='auto'):
    n = len(self.links)

    if q is None:
        q = self.q

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_number(tol) or tol <= 0:
        raise Exception("The parameter 'tol' should be a positive number.")

    if not Utils.is_a_number(max_dist) or tol <= 0:
        raise Exception("The parameter 'max_dist' should be a positive number, or np.inf.")

    if not Utils.is_a_natural_number(no_iter_max) or no_iter_max <= 0:
        raise Exception("The parameter 'no_iter_max' should be a positive natural number.")

    if not (old_dist_struct is None):
        try:
            if not id(old_dist_struct.robot) == id(self):
                Exception("The parameter 'old_dist_struct' is a 'DistStructRobotAuto' object, but it " \
                          "must have to be relative to the SAME robot object, and " \
                          "this is not the case.")
        except:
            raise Exception("The parameter 'old_dist_struct' must be a 'DistStructRobotAuto' object.")

    if not Utils.is_a_number(h) or h < 0:
        raise Exception("The optional parameter 'h' must be a nonnegative number.")

    if not Utils.is_a_number(eps) or eps < 0:
        raise Exception("The optional parameter 'eps' must be a nonnegative number.")
    
    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        if h >0 or eps > 0:
            raise Exception("In Python mode, smoothing parameters 'h' and 'eps' must be set to 0!")
         
        return _compute_dist_auto_python(self, Utils.cvt(q), old_dist_struct, tol, no_iter_max, max_dist)
    else:
        

        old_dsra = _diststructrobotauto_py2cpp(old_dist_struct)       
        new_dsra = self.cpp_robot.compute_dist_auto(Utils.cvt(q), old_dsra, tol, no_iter_max, max_dist, h+1e-6, eps+1e-6)

        return _diststructrobotauto_cpp2py(new_dsra, self)


def _compute_dist_auto_python(self, q=None, old_dist_struct=None, tol=0.0005, no_iter_max=20, max_dist = np.inf):
    n = len(self.links)

    dist_struct = DistStructRobotAuto(self)

    jac_dh, mth_dh = self.jac_geo(q, "dh", mode = "python")

    col_object_copy = []

    # Update all collision objects of all links
    for i in range(n):
        col_object_copy.append([])
        for j in range(len(self.links[i].col_objects)):
            temp_copy = self.links[i].col_objects[j][0].copy()
            htmd = self.links[i].col_objects[j][1]
            temp_copy.set_ani_frame(mth_dh[i][:, :] * htmd)
            col_object_copy[i].append(temp_copy)

    # Compute the distance structure
    for i in range(n):
        for j in range(i+2,n):
            for isub in range(len(self.links[i].col_objects)):
                for jsub in range(len(self.links[j].col_objects)):

                    est_dist = 0 if math.isinf(max_dist) else Utils.compute_aabbdist(col_object_copy[i][isub], col_object_copy[j][jsub])

                    if est_dist <= max_dist:

                        if old_dist_struct is None:
                            p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))
                        else:
                            p_obj_0 = old_dist_struct.get_item(i, isub, j, jsub).point_object

                        p_obj_i, p_obj_j, d = Utils.compute_dist(col_object_copy[i][isub], col_object_copy[j][jsub] \
                                                                 , p_obj_0, tol, no_iter_max)

                        jac_obj_i = jac_dh[i][0:3, :] - Utils.S(p_obj_i - mth_dh[i][0:3, 3]) * jac_dh[i][3:6, :]
                        jac_obj_j = jac_dh[j][0:3, :] - Utils.S(p_obj_j - mth_dh[j][0:3, 3]) * jac_dh[j][3:6, :]

                        jac_dist = (p_obj_i - p_obj_j).T * (jac_obj_i - jac_obj_j) / d
                        dist_struct._append(i, isub, j, jsub, d, p_obj_i, p_obj_j, jac_dist)



    return dist_struct



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_magician_e6(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")


    c=0

    link_info = [[0, 0, 0, 0, 0, 0], #"theta" rotation in z
                 [0.167, -0.086, 0.086, -0.086, 0.101, -0.065],  # "d" translation in z
                 [-np.pi/2, 0, 0, np.pi/2, -np.pi/2, np.pi],  # "alfa" rotation in x
                 [0, 0.189, 0.16, 0, 0, 0],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0]] #joint type

    n = 6
    scale = 1

    # Collision model
    col_model = [[], [], [], [], [], []]


    col_model[0].append(Cylinder(htm=Utils.trn([0, 0.05, 0])*Utils.rotx(np.pi/2),
                                 name=name + "_C0_0", radius=0.045, height=0.175, color="red", opacity=0.3))
    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, -0.04]),
                                 name=name + "_C0_1", radius=0.045, height=0.175, color="red", opacity=0.3))
    col_model[1].append(Box(htm=Utils.trn([-0.07, 0, -0.005]),
                            name=name + "_C1_0", width=0.18, depth=0.09, height=0.06, color="blue", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, 0.05]),
                                 name=name + "_C1_1", radius=0.045, height=0.175, color="blue", opacity=0.3))
    col_model[2].append(Box(htm=Utils.trn([-0.07, 0, 0.01]),
                            name=name + "_C2_0", width=0.18, depth=0.09, height=0.06, color="green", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, -0.04]),
                                 name=name + "_C2_1", radius=0.04, height=0.175, color="green", opacity=0.3))
    col_model[3].append(Cylinder(htm=Utils.trn([0, 0, 0.05]),
                                 name=name + "_C3_0", radius=0.035, height=0.13, color="yellow", opacity=0.3))
    col_model[4].append(Cylinder(htm=Utils.trn([0, 0, 0]),
                                 name=name + "_C3_1", radius=0.035, height=0.14, color="cyan", opacity=0.3))

    # Create 3d objects
    

    htm1 = np.matrix([[-0., -1.,  0., -0.    ],[-0.,  0., -1.,  0.0402],[ 1., -0., -0.,  0.    ],[ 0.,  0.,  0.,  1.    ]])
    htm2 = np.matrix([[-0.,  1., -0., -0.1892],[ 1.,  0., -0.,  0.    ],[-0., -0., -1.,  0.04  ],[ 0.,  0., 0.,  1.    ]])
    htm3 = np.matrix([[-0.,  1., -0., -0.1601],[ 1.,  0., -0., -0.    ],[-0., -0., -1., -0.049 ],[ 0.,  0.,  0.,  1.    ]])
    htm4 = np.matrix([[ 1., -0., -0., -0.    ],[-0., -0., -1.,  0.032 ],[ 0.,  1., -0., -0.0001],[ 0.,  0.,  0.,  1.    ]])
    htm5 = np.matrix([[-0., -1., -0., -0.,    ],[-0.,  0., -1.,  0.0341],[ 1., -0., -0., -0.    ],[ 0.,  0.,  0.,  1.    ]])
    htm6 = np.matrix([[ 1., -0., -0., -0.    ],[ 0.,  1., -0., -0.0001],[ 0.,  0.,  1., -0.018 ],[ 0.,  0.,  0.,  1.    ]])
    

    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/base_link.stl',
        scale,
        np.identity(4),
        MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                     color=color, opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link1.stl',
            scale,
            htm1,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link2.stl',
            scale,
            htm2,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link3.stl',
            scale,
            htm3,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link4.stl',
            scale,
            htm4,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link5.stl',
            scale,
            htm5,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#C0C0C0",
                         opacity=opacity))
         ]
    )
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/MagicianE6/Link6.stl',
            scale,
            htm6,
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="black",
                         opacity=opacity))
         ]
    )

    # Create links
    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    htm_base_0 = Utils.trn([0, 0, 0])
    q0 = [-np.pi/2, -np.pi/2, 0.0, np.pi/2, 0.0, 0.0]

    #Create joint limits
    joint_limits = np.matrix([[-2.5*np.pi, 1.5*np.pi],[-np.pi/2-2.356,-np.pi/2+2.356],[-2.6878, 2.6878],[-2.7925+np.pi/2,2.7925+np.pi/2],[-3.0194,3.0194],[-6.28,6.28]])

    return base_3d_obj, links, htm_base_0, np.identity(4), q0, joint_limits



from utils import Utils

from graphics.meshmaterial import MeshMaterial
from graphics.model3d import Model3D
from simobjects.rigidobject import RigidObject

import numpy as np


def _create_davinci_chest(name, color, opacity):

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    link_info = np.array([
        # "theta" rotation in z
        [0, 0, 0, 0,        0,       0, 0],
        # "d" translation in z
        [0, 0, 0, 0,        0,       0, 0],
        # "alfa" rotation in x
        [0, 0, 0, 0, -np.pi/2, np.pi/2, 0],
        # "a" translation in x
        [0, 0, 0, 0,        0,       0, 0],
        # joint type
        [0, 0, 0, 0,        0,       1, 1]
    ])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    chest_obj = []
    mesh = MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5,
                        normal_scale=[0.5, 0.5], color=color,
                        opacity=opacity, side="DoubleSide")
    # original model is rotated (Robot fron = plane X x Y)

    Q01 = Utils.rotx(-np.pi/2) * Utils.rotz(np.pi) * Utils.rotz(link_info[0, 0]) * Utils.trn([0, 0, link_info[1, 0]]) * Utils.rotx(link_info[2, 0]) * Utils.trn(
        [link_info[3, 0], 0, 0])
    Q12 = Q01 * (Utils.rotz(link_info[0, 1]) * Utils.trn([0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn(
        [link_info[3, 1], 0, 0]))
    Q23 = Q12 * (Utils.rotz(link_info[0, 2]) * Utils.trn([0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn(
        [link_info[3, 2], 0, 0]))
    Q34 = Q23 * (Utils.rotz(link_info[0, 3]) * Utils.trn([0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn(
        [link_info[3, 3], 0, 0]))
    Q45 = Q34 * (Utils.rotz(link_info[0, 4]) * Utils.trn([0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn(
        [link_info[3, 4], 0, 0]))
    Q56 = Q45 * (Utils.rotz(link_info[0, 5]) * Utils.trn([0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn(
        [link_info[3, 5], 0, 0]))
    Q67 = Q56 * (Utils.rotz(link_info[0, 6]) * Utils.trn([0, 0, link_info[1, 6]]) * Utils.rotx(link_info[2, 6]) * Utils.trn(
        [link_info[3, 6], 0, 0]))

    link1_mth = Utils.inv_htm(Q01)
    chest_obj.extend([
        # feet
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/1.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
    ])

    link2_mth = Utils.inv_htm(Q12)
    chest_obj.extend([
        # base rectangle
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/2.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
        # vertical tower
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/3.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/17.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
        # short cable
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/18.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
    ])

    chest = RigidObject(chest_obj, name + "_chest")

    return chest



import numpy as np
from utils import *
from graphics.model3d import *


class Link:
    """
  A class that contains the information of the links of the robot.


  Parameters
  ----------

  joint_number : positive int
      The joint number in the kinematic chain (0 is the first joint).

  theta : float
      The 'theta' parameter in the Denavit-Hartenberg convention (rotation in z), in rad.

  d : float
      The 'd' parameter in the Denavit-Hartenberg convention (displacement in z), in meters.

  alpha : float
      The 'alpha' parameter in the Denavit-Hartenberg convention (rotation x), in rad.

  a : float
      The 'a' parameter in the Denavit-Hartenberg convention (displacement in x), in meters.

  joint_type : 0 or 1
      The joint type. "0" is rotative and "1" is prismatic.

  list_model_3d : A list of 'uaibot.Model3D' objects
      The 3d models that compose the links.

  """

    #######################################
    # Attributes
    #######################################

    @property
    def theta(self):
        """The 'theta' parameter of the Denavit-Hartenberg convention (in rad)"""
        return self._theta

    @property
    def d(self):
        """The 'd' parameter of the Denavit-Hartenberg convention (in meters)"""
        return self._d

    @property
    def a(self):
        """The 'a' parameter of the Denavit-Hartenberg convention (in meters)"""
        return self._a

    @property
    def alpha(self):
        """The 'alpha' parameter of the Denavit-Hartenberg convention (in rad)"""
        return self._alpha

    @property
    def joint_number(self):
        """The joint number in the kinematic chain."""
        return self._joint_number

    @property
    def joint_type(self):
        """The joint type (0=revolute, 1=prismatic)."""
        return self._joint_type

    @property
    def col_objects(self):
        """Collection of objects that compose the collision model of the link."""
        return self._col_objects

    @property
    def list_model_3d(self):
        """The list of 3d models of the object."""
        return self._list_model_3d

    @property
    def show_frame(self):
        """If the Denavit-Hartenberg frame of the link is displayed in simulation."""
        return self._show_frame

    #######################################
    # Constructor
    #######################################

    def __init__(self, joint_number, theta, d, alpha, a, joint_type, list_model_3d, show_frame=False):

        # Error handling
        if str(type(joint_number)) != "<class 'int'>" or joint_number < 0:
            raise Exception("The 'joint_number' parameter should be a nonnegative integer.")

        if not Utils.is_a_number(theta):
            raise Exception("The 'theta' parameter should be a float.")

        if not Utils.is_a_number(d):
            raise Exception("The 'd' parameter should be a float.")

        if not Utils.is_a_number(alpha):
            raise Exception("The 'alpha' parameter should be a float.")

        if not Utils.is_a_number(a):
            raise Exception("The 'a' parameter should be a float.")

        if joint_type != 0 and joint_type != 1:
            raise Exception("The 'joint_type' parameter should be either '0' (rotative) or '1' (prismatic).")

        if not (str(type(list_model_3d)) == "<class 'list'>"):
            raise Exception("The parameter 'list_model_3d' should be a list of 'uaibot.Model3D' objects.")
        else:
            for i in range(len(list_model_3d)):
                if not (Utils.get_uaibot_type(list_model_3d[i]) == "uaibot.Model3D"):
                    raise Exception(
                        "The parameter 'list_model_3d' should be a list of 'uaibot.Model3D' objects.")

        if not str(type(show_frame)) == "<class 'bool'>":
            raise Exception("The parameter 'show_frame' should be a boolean.")

        # Code
        self._joint_number = joint_number
        self._theta = theta
        self._d = d
        self._alpha = alpha
        self._a = a
        self._joint_type = joint_type
        self._col_objects = []
        self._list_model_3d = list_model_3d
        self._show_frame = show_frame

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Link " + str(self._joint_number) + "' "

        if self._joint_type == 0:
            string += "with rotative joint:\n\n"
            string += " θ (rad) : [variable] \n"
            string += " d (m)   : " + str(self._d) + " \n"
        if self._joint_type == 1:
            string += "with prismatic joint:\n\n"
            string += " θ (rad) : " + str(self._theta) + " \n"
            string += " d (m)   : [variable] \n"

        string += " α (rad) : " + str(self._alpha) + " \n"
        string += " a (m)   : " + str(self._a) + " \n"

        return string

    #######################################
    # Methods
    #######################################

    def attach_col_object(self, obj, htm):
        """
    Attach an object (ball, box or cylinder) into the link as a collision
    object.
    
    Parameters
    ----------
    obj: object
        Object to be added.

    htm : 4x4 numpy array
        The transformation between the link's HTM and the object's HTM

    """

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not (Utils.is_a_simple_object(obj)):
            raise Exception("The parameter 'obj' must be one of the following types: " + str(Utils.IS_SIMPLE) + ".")

        self._col_objects.append([obj, htm])

    def gen_code(self, name):
        """Generate code for injection."""

        name_link = "link_" + str(self.joint_number) + "_" + name

        string = "const object3d_" + name_link + "_list = [];\n\n"

        for i in range(len(self._list_model_3d)):
            string += self.list_model_3d[i].gen_code(name_link + "_obj_" + str(i))
            string += "object3d_" + name_link + "_list.push(object3d_" + name_link + "_obj_" + str(i) + ");\n"

        string += "\n"
        string += "const " + name_link + " = {\n"
        string += "theta: " + str(self.theta) + ", \n"
        string += "d: " + str(self.d) + ", \n"
        string += "a: " + str(self.a) + ", \n"
        string += "alpha: " + str(self.alpha) + ", \n"
        string += "jointType: " + str(self.joint_type) + ", \n"
        string += "showFrame: " + ("true" if self.show_frame else "false") + ",\n"
        string += "model3d: object3d_" + name_link + "_list}; \n \n"

        return string



from utils import *


# Set config. Restart animation queue
def _set_ani_frame(self, q=None, htm=None, enforce_joint_limits=False):
    if q is None:
        q = self.q0

    if htm is None:
        htm = self.htm

    # Error handling
    if not Utils.is_a_vector(q, len(self.links)):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not str(type(enforce_joint_limits)) == "<class 'bool'>":
        raise Exception("The parameter 'enforce_joint_limits' must be a boolean.")
    # end error handling

    self._frames = []
    self.code = ''
    self._max_time = 0
    self.add_ani_frame(0, q, htm, enforce_joint_limits)
    
    if self.eef_frame_visible:
        self._eef_frame.add_ani_frame(0,self.fkm(q))



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *
from simobjects.rigidobject import *
from simobjects.group import *

from .links import *

from robot._create_darwin_mini_arm import _create_darwin_mini_arm
from robot._create_darwin_mini_leg_left import _create_darwin_mini_leg_left
from robot._create_darwin_mini_leg_right import _create_darwin_mini_leg_right




def _create_darwin_mini(htm, name, color, opacity):

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    param_arm_left = _create_darwin_mini_arm(np.identity(4), "arm_left", color, 1)
    param_arm_right = _create_darwin_mini_arm(np.identity(4), "arm_right", color, 1)
    param_leg_left = _create_darwin_mini_leg_left(np.identity(4), "leg_left", color, 1)
    param_leg_right = _create_darwin_mini_leg_right(np.identity(4), "leg_right", color, 1)

    chestpart1 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/darwin_chest.obj',
        0.004,
        htm * Utils.trn([-0.3, -0.179, 0.37]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                     opacity=opacity, side="DoubleSide"))
    chestpart2 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/frm_body_01.obj',
        0.004,
        htm * Utils.trn([-0.3, -0.179, 0.37]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                     opacity=opacity, side="DoubleSide"))
    chestpart3 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320.obj',
        0.004,
        htm * Utils.trn([0.265, 0.405, 0.32]) * Utils.rotz(-3.14 / 2) * Utils.rotx(3.14 / 3 + 0.25) * Utils.rotz(
            3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=1, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chestpart4 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320.obj',
        0.004,
        htm * Utils.trn([0.265, -0.405, 1.305]) * Utils.rotx(3.14) * Utils.rotz(-3.14 / 2) * Utils.rotx(
            3.14 / 3 + 0.25) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=1, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chestpart5 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPU-2.obj',
        0.004,
        htm * Utils.trn([0.05, 0.13, -0.18]) * Utils.rotz(3.14 / 2) * Utils.trn([-0.42, -0.179, 0.575]) * Utils.rotz(
            3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chestpart6 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
        0.004,
        htm * Utils.trn([0.69, 0.1, 0.475]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
            [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chestpart7 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/SPU-2.obj',
        0.004,
        htm * Utils.trn([0.05, -0.06, -0.18]) * Utils.rotz(3.14 / 2) * Utils.trn([-0.42, -0.179, 0.575]) * Utils.rotz(
            3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chestpart8 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
        0.004,
        htm * Utils.trn([0.69, -0.09, 0.475]) * Utils.rotx(0.08) * Utils.roty(-3.14 / 2) * Utils.trn(
            [-0.3, -0.179, 0.15]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    chest = RigidObject(
        [chestpart1, chestpart2, chestpart3, chestpart4, chestpart5, chestpart6, chestpart7, chestpart8],
        name + "_chest")

    headpart1 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/darwin_head.obj',
        0.004,
        htm * Utils.trn([-0.3, -0.179, 0.37]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                     opacity=opacity, side="DoubleSide"))

    headpart2 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/eye.obj',
        0.0009,
        htm * Utils.trn([0.177, -0.043, 0.982]) * Utils.rotz(3.14 / 2 - 0.3),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="blue",
                     opacity=opacity, side="DoubleSide"))
    headpart3 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/eye.obj',
        0.0009,
        htm * Utils.trn([0.171, 0.053, 0.982]) * Utils.rotz(3.14 / 2 + 0.4),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="blue",
                     opacity=opacity, side="DoubleSide"))
    headpart4 = Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/DarwinMini/XL-320-2.obj',
        0.004,
        htm * Utils.trn([-0.3, -0.179, 0.37]) * Utils.rotz(3.14 / 2) * Utils.rotx(3.14 / 2),
        MeshMaterial(metalness=0.8, clearcoat=0.5, roughness=0.5, normal_scale=[0.5, 0.5], color="#302b2b",
                     opacity=opacity, side="DoubleSide"))

    head = RigidObject([headpart1, headpart2, headpart3, headpart4], name + "_head")

    return param_arm_left, param_arm_right, param_leg_left, param_leg_right, head, chest



from utils import *


def _attach_object(self, obj):
    # Error handling

    if not (Utils.get_uaibot_type(obj) in Utils.IS_GROUPABLE):
        raise Exception("The parameter 'obj' must be one of the following types: " + str(Utils.IS_GROUPABLE) + ".")
    # end error handling

    self._attached_objects.append([obj, Utils.inv_htm(self.fkm()) @ obj.htm])



from utils import *
import numpy as np
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject
from typing import Optional, Tuple, List

from uaibot.graphics.model3d  import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from uaibot.graphics.model3d import Model3D


class RigidObject:
    """
   A rigid object.

   Parameters
   ----------

   name : string
       The object's name.
       (default: '' (automatic)).

   list_model_3d : list of 'Model3D' objects
       The 3d model that compose the object.

   htm : 4x4 numpy matrix
       The object's configuration.
       (default: the same as the current HTM).

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @property
    def htm(self) -> "HTMatrix":
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def list_model_3d(self) -> List["Model3D"]:
        """The list of 3d models of the object."""
        return self._list_model_3d

    #######################################
    # Constructor
    #######################################

    def __init__(self, list_model_3d: List["Model3D"], name: str ="", 
                 htm: "HTMatrix"=np.identity(4)) -> "RigidObject":

        # Error handling

        if name=="":
            name="var_rigidobject_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. "
                "It should not begin with a number.")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not (str(type(list_model_3d)) == "<class 'list'>"):
            raise Exception("The parameter 'list_model_3d' should be a list of 'uaibot.Model3D' objects.")
        else:
            for i in range(len(list_model_3d)):
                if not (Utils.get_uaibot_type(list_model_3d[i]) == "uaibot.Model3D"):
                    raise Exception(
                        "The parameter 'list_model_3d' should be a list of 'uaibot.Model3D' objects.")

        # end error handling

        self._htm = np.matrix(htm)
        self._name = name
        self._list_model_3d = list_model_3d
        self._max_time = 0

        self.set_ani_frame(self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Rigid Object with name '" + self.name + "': \n\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, htm: Optional["HTMatrix"]=None) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy array
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        f = [time, np.around(htm[0,0],4).item(), np.around(htm[0,1],4).item(), np.around(htm[0,2],4).item(), np.around(htm[0,3],4).item(),
             np.around(htm[1,0],4).item(), np.around(htm[1,1],4).item(), np.around(htm[1,2],4).item(), np.around(htm[1,3],4).item(),
             np.around(htm[2,0],4).item(), np.around(htm[2,1],4).item(), np.around(htm[2,2],4).item(), np.around(htm[2,3],4).item(),
             0, 0, 0, 1]

        self._htm = htm
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    # Set config. Restart animation queue
    def set_ani_frame(self, htm: Optional["HTMatrix"]=None) -> None:
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE RIGID OBJECT '" + self.name + "'\n\n"

        string += "var list_object3d_" + self.name + " = [];\n"
        for i in range(len(self.list_model_3d)):
            string += self.list_model_3d[i].gen_code(self.name + "_" + str(i)) + ";\n"
            string += "list_object3d_" + self.name + ".push(object3d_" + self.name + "_" + str(i) + ");\n"

        string += "const var_" + self.name + " = new RigidObject(" + str(
            self._frames) + ", list_object3d_" + self.name + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string
